"use strict";
/**
 * @license
 * Copyright 2018 Google LLC
 *
 * Use of this source code is governed by an MIT-style
 * license that can be found in the LICENSE file or at
 * https://opensource.org/licenses/MIT.
 * =============================================================================
 */
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * Testing utilities.
 */
var tfjs_core_1 = require("@tensorflow/tfjs-core");
// tslint:disable-next-line: no-imports-from-dist
var jasmine_util_1 = require("@tensorflow/tfjs-core/dist/jasmine_util");
var errors_1 = require("../errors");
// Register backends.
jasmine_util_1.registerTestEnv({ name: 'cpu', backendName: 'cpu' });
jasmine_util_1.registerTestEnv({
    name: 'webgl2',
    backendName: 'webgl',
    flags: {
        'WEBGL_VERSION': 2,
        'WEBGL_CPU_FORWARD': false,
        'WEBGL_SIZE_UPLOAD_UNIFORM': 0
    }
});
/**
 * Expect values are close between an Tensor or number array.
 * @param actual
 * @param expected
 */
function expectTensorsClose(actual, expected, epsilon) {
    if (actual == null) {
        throw new errors_1.ValueError('First argument to expectTensorsClose() is not defined.');
    }
    if (expected == null) {
        throw new errors_1.ValueError('Second argument to expectTensorsClose() is not defined.');
    }
    if (actual instanceof tfjs_core_1.Tensor && expected instanceof tfjs_core_1.Tensor) {
        if (actual.dtype !== expected.dtype) {
            throw new Error("Data types do not match. Actual: '" + actual.dtype + "'. " +
                ("Expected: '" + expected.dtype + "'"));
        }
        if (!tfjs_core_1.util.arraysEqual(actual.shape, expected.shape)) {
            throw new Error("Shapes do not match. Actual: [" + actual.shape + "]. " +
                ("Expected: [" + expected.shape + "]."));
        }
    }
    var actualData = actual instanceof tfjs_core_1.Tensor ? actual.dataSync() : actual;
    var expectedData = expected instanceof tfjs_core_1.Tensor ? expected.dataSync() : expected;
    tfjs_core_1.test_util.expectArraysClose(actualData, expectedData, epsilon);
}
exports.expectTensorsClose = expectTensorsClose;
/**
 * Expect values in array are within a specified range, boundaries inclusive.
 * @param actual
 * @param expected
 */
function expectTensorsValuesInRange(actual, low, high) {
    if (actual == null) {
        throw new errors_1.ValueError('First argument to expectTensorsClose() is not defined.');
    }
    tfjs_core_1.test_util.expectValuesInRange(actual.dataSync(), low, high);
}
exports.expectTensorsValuesInRange = expectTensorsValuesInRange;
/**
 * Describe tests to be run on CPU and GPU.
 * @param testName
 * @param tests
 */
function describeMathCPUAndGPU(testName, tests) {
    jasmine_util_1.describeWithFlags(testName, jasmine_util_1.ALL_ENVS, function () {
        tests();
    });
}
exports.describeMathCPUAndGPU = describeMathCPUAndGPU;
/**
 * Describe tests to be run on CPU only.
 * @param testName
 * @param tests
 */
function describeMathCPU(testName, tests) {
    jasmine_util_1.describeWithFlags(testName, { predicate: function (testEnv) { return testEnv.backendName === 'cpu'; } }, function () {
        tests();
    });
}
exports.describeMathCPU = describeMathCPU;
/**
 * Describe tests to be run on GPU only.
 * @param testName
 * @param tests
 */
function describeMathGPU(testName, tests) {
    jasmine_util_1.describeWithFlags(testName, { predicate: function (testEnv) { return testEnv.backendName === 'webgl'; } }, function () {
        tests();
    });
}
exports.describeMathGPU = describeMathGPU;
/**
 * Check that a function only generates the expected number of new Tensors.
 *
 * The test  function is called twice, once to prime any regular constants and
 * once to ensure that additional copies aren't created/tensors aren't leaked.
 *
 * @param testFunc A fully curried (zero arg) version of the function to test.
 * @param numNewTensors The expected number of new Tensors that should exist.
 */
function expectNoLeakedTensors(
// tslint:disable-next-line:no-any
testFunc, numNewTensors) {
    testFunc();
    var numTensorsBefore = tfjs_core_1.memory().numTensors;
    testFunc();
    var numTensorsAfter = tfjs_core_1.memory().numTensors;
    var actualNewTensors = numTensorsAfter - numTensorsBefore;
    if (actualNewTensors !== numNewTensors) {
        throw new errors_1.ValueError("Created an unexpected number of new " +
            ("Tensors.  Expected: " + numNewTensors + ", created : " + actualNewTensors + ". ") +
            "Please investigate the discrepency and/or use tidy.");
    }
}
exports.expectNoLeakedTensors = expectNoLeakedTensors;
//# sourceMappingURL=test_utils.js.map